/*******************************************************************************
 * Copyright (C) 2024
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
#include "NetworkInterfaceTlm.h"
#include <iostream>

#include "ratatoskrUtils/traffic/Flit.h"
#include "utils/configuration.h"
// #include "utils/utils.h"

//DECLARE_EXTENDED_PHASE(INTERNAL_PROC_PHASE);

using namespace std;

NetworkInterfaceTlm::NetworkInterfaceTlm(sc_module_name nm, Node& node) : 
            NetworkInterface(nm, node), lastReceivedCreditID(-1), credit_counter(0), ni_name(nm), resp_in_progress(false)
             {
                target_socket.register_nb_transport_fw(this, &NetworkInterfaceTlm::nb_transport_fw); ////from core
                //sc_report_handler::set_actions(NI_LOG, SC_INFO, SC_LOG|SC_DISPLAY);
                sc_report_handler::set_log_file_name("out/report.log"); //from core
                sc_report_handler::set_actions(LOG_NAME, SC_INFO, SC_LOG|SC_DISPLAY); //from core
    
    try {
        this->id = source_core_id;
        this->dbid = rep.registerElement("ProcessingElement", this->id);
        this->node = node;
        this->packetPortContainer = new PacketPortContainer(
                ("NI_PACKET_CONTAINER"+std::to_string(this->id)).c_str());
        //copy(max_pos, max_pos + 3, this->max_pos);
    }
    catch (exception& e) {
        log_error("Initialization Error: " + string(e.what()));
    }
    SC_THREAD(thread);
    sensitive << clk.pos() << clk.neg();

    SC_METHOD(receivePacketFromPE);
    sensitive << packetPortContainer->portValidIn.pos();

}

NetworkInterfaceTlm::~NetworkInterfaceTlm() {
    delete packetPortContainer;
}

void NetworkInterfaceTlm::initialize(){
    /* NOT IMPLEMENTED */
}

void NetworkInterfaceTlm::bind(Connection* con, SignalContainer* sigContIn, SignalContainer* sigContOut) {
    if (con==nullptr)
        packetPortContainer->bind(sigContIn, sigContOut);
}

void NetworkInterfaceTlm::receiveFlitFromRouter(){
    /* NOT IMPLEMENTED */
}

void NetworkInterfaceTlm::receivePacketFromPE() {
    if (packetPortContainer->portValidIn.posedge()) {
        //LOG(globalReport.verbose_pe_function_calls, "NI" << this->id << "(Node" << node.id << ")\t\t- receive()");
        Packet* packet = packetPortContainer->portDataIn.read();
        generateFlitsForPacket(packet);
        packet_send_queue.push(packet);
    }
}

void NetworkInterfaceTlm::getDestination(Packet* p, std::string pos[2]){
    pos[0] = to_string(node.pos.x) + "," +
            to_string(node.pos.y) + "," + to_string(node.pos.z);
    pos[1] = to_string(p->dst.pos.x) + "," +
            to_string(p->dst.pos.y) + "," +
            to_string(p->dst.pos.z);
}

void NetworkInterfaceTlm::generateFlitsForPacket(Packet* p) {
    int flitsPerPacket = p->size;
    for (int i = 0; i<flitsPerPacket; ++i) {
        FlitType flitType;
        if (flitsPerPacket==1)
            flitType = SINGLE;
        else if (i%flitsPerPacket==0)
            flitType = HEAD;
        else if (i%flitsPerPacket==flitsPerPacket-1)
            flitType = TAIL;
        else
            flitType = BODY;
        int seqNum = i%flitsPerPacket;
        Flit* current_flit = new Flit(flitType, seqNum, p, p->dataType, sc_time_stamp().to_double());
        p->toTransmit.push_back(current_flit->id);
        p->flits.push_back(current_flit);
        
        std::string start_end_pos[2];
        getDestination(p, start_end_pos);
        log_info("Flits generated for packet of id "+ to_string(p->id)+
         " to be sent to "+start_end_pos[1]+" from "+ start_end_pos[0]);
    }
}

void NetworkInterfaceTlm::send_packet_to_router() {
 // if no flits to transmit, generate them
    Packet* p = packet_send_queue.front();
    if (p->toTransmit.empty()){
        generateFlitsForPacket(p);
    }
    // get current flit to send
    flitID_t f_id = p->toTransmit.front();
    auto iter = find_if(p->flits.begin(), p->flits.end(),
            [&f_id](Flit* f) { return f->id==f_id; });
    Flit* current_flit = *iter;
    current_flit->injectionTime = sc_time_stamp().to_double();

    auto toDelete_pos = find(p->toTransmit.begin(), p->toTransmit.end(), current_flit->id);
    p->toTransmit.erase(toDelete_pos);
    p->inTransmit.push_back(current_flit->id);

    if (p->toTransmit.empty()) {
        packet_send_queue.pop();
    }
    
    my_payload* trans = nullptr;

    unsigned int src_core_id = p->src.id;
    unsigned int src_router_id = src_core_id/CORE_NO; 
    unsigned int dst_core_id = p->dst.id; 
    unsigned int destination_router_id = dst_core_id/CORE_NO;
    sc_time delay = sc_time(0, SC_NS);

    trans = new my_payload(src_core_id, src_router_id, dst_core_id, destination_router_id, current_flit, sc_time_stamp());
    // Construct the address (calculation based on router/core IDs)
    unsigned int offset = 0x20;  // offset
    unsigned int address = (destination_router_id << 12) | (dst_core_id << 8) | offset;
    trans->set_address(address);
    log_info( "Network Interface " + std::to_string(src_core_id) + " of Source Router: " + std::to_string(source_router_id) 
                + ": Processing packet flit with ID: " + std::to_string(trans->data->id) + " for router " + std::to_string(destination_router_id) 
                + " and Network Interface " + std::to_string(dst_core_id) + " at time: " + trans->timestamp.to_string());
            
            
    // Set up the rest of the transaction
    trans->set_data_length(p->size);
    trans->set_streaming_width(4);
    trans->set_byte_enable_ptr(nullptr);
    trans->set_dmi_allowed(false);
    trans->set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    tlm::tlm_phase phase = tlm::BEGIN_REQ; // Transaction phase
    tlm::tlm_sync_enum status = socket->nb_transport_fw(*trans, phase, delay); // Non-blocking transport

   
    bool retry = false;
    do {
        if (!retry) {
            tlm::tlm_sync_enum status = socket->nb_transport_fw(*trans, phase, delay); // Non-blocking transport

            if (status == tlm::TLM_COMPLETED || phase == tlm::END_REQ) {
                    log_info("Network Interface " + ni_name + ": Transaction successful, flit sent ID: " + std::to_string(trans->data->id) + " at time: " + sc_time_stamp().to_string());
                    delete trans; // Clean up the transaction
                    retry = false;  // Move on to the next transaction
                    //wait(50, SC_NS);
                } else if (status == tlm::TLM_ACCEPTED) {
                    log_info("Network Interface " + ni_name + ": Transaction accepted, waiting for completion.");
                    wait(0, SC_NS);  // Wait for response
                    retry = false;  // No need to retry, accepted transaction will continue
                } else if (status == tlm::TLM_UPDATED && phase == tlm::BEGIN_REQ) {
                    log_info("Network Interface " + ni_name + ": FIFO full, retrying...");
                    wait(delay + sc_time(20, SC_NS));  // Add some delay before retrying
                    retry = true;  // Retry the same transaction
                }
        } 

        
    
    }while(retry);

    wait(0,SC_NS);    
}


void NetworkInterfaceTlm::thread() {
    while(true){
        //log_info("send_data_process()");
        if (clk.posedge()) {
            if (!packet_send_queue.empty()) {
                //if (credit_counter != 0){
                    send_packet_to_router();
                //}
                //else {
                //   log_info("Waiting for Router!");
                //}
            }
            if (!packet_recv_queue.empty()) {
                if (packetPortContainer->portFlowControlIn.read()) {
                    Packet* p = packet_recv_queue.front();
                    packet_recv_queue.pop();
                    packetPortContainer->portValidOut.write(true);
                    packetPortContainer->portDataOut.write(p);
                    // log arrival of packet
                    std::string start_end_pos[2];
                    getDestination(p, start_end_pos);
                    log_info("Packet with id " + to_string(p->id) + 
                               " arrived at PE "+start_end_pos[1]);
                }
            }
        }
        else if (clk.negedge()) {
            packetPortContainer->portValidOut.write(false);
        }
        wait();
    }
}

tlm::tlm_sync_enum NetworkInterfaceTlm::nb_transport_fw(tlm::tlm_generic_payload& trans, tlm::tlm_phase& phase, sc_time& delay) {
        if (phase == tlm::BEGIN_REQ) {
        my_payload* my_trans = dynamic_cast<my_payload*>(&trans);

        // Print which core is processing the received data
        log_info("Network Interface " + std::to_string(source_core_id) + " of parent Router: " + std::to_string(source_router_id) 
                + ": Processing  transaction " + "and received flit ID: " + std::to_string(my_trans->data->id));
            
        // Send acknowledgment or status back 
        trans.set_response_status(tlm::TLM_OK_RESPONSE);

        // Log the received data and remove the payload
        delete my_trans;
        return tlm::TLM_COMPLETED;
    }
    return tlm::TLM_ACCEPTED;
    }


void NetworkInterfaceTlm::receive_and_process_flit(tlm::tlm_generic_payload& trans){
    trans.set_response_status(tlm::TLM_OK_RESPONSE);
    unsigned char* data_ptr = trans.get_data_ptr();
    Flit* received_flit = reinterpret_cast<Flit*>(data_ptr);
    Packet* p = received_flit->packet;
    double time = sc_time_stamp().to_double();

    auto position = find(p->inTransmit.begin(), p->inTransmit.end(), received_flit->id);
    if (position!=p->inTransmit.end())
        p->inTransmit.erase(position);
    p->transmitted.push_back(received_flit->id);

    if (received_flit->type==TAIL || received_flit->type==SINGLE){
        stringstream ss;
        std::string flit_type = received_flit->type==SINGLE ? "Single":"Tail";
        ss << "Receive Flit " << flit_type << *received_flit;
        log_info(ss.str());
    }
    if (!p->toTransmit.empty() || !p->inTransmit.empty()) {
        stringstream ss;
        ss << "Received Tail Flit, but still missing flits! "
                << *received_flit;
        log_info(ss.str());
    }

    globalReport.issueNoCOutputDataAmount(sc_time_stamp(),globalResources.bitWidth);
    if (p->toTransmit.empty() && p->inTransmit.empty())
        packet_recv_queue.push(p);
    //credit_counter++;

    // if(resp_in_progress) {
    //     if(nxt_resp_pend){
    //        // log_fatal("Attempt to have two pending responses in target");
    //         nxt_resp_pend = &trans;
    //     }
    // }
    //else{ send_response(trans); }
}




void NetworkInterfaceTlm::log_info(std::string msg){
    // string log_msg = ni_name + ": (Node" + 
    //                     to_string(node.id)+"): "+ msg;
    SC_REPORT_INFO(LOG_NAME, (msg).c_str());
}

void NetworkInterfaceTlm::log_error(std::string msg){
    // string log_msg = ni_name + ": (Node" + 
    //                     to_string(node.id)+"): "+ msg;
    SC_REPORT_ERROR(LOG_NAME, (msg).c_str());
}

void NetworkInterfaceTlm::log_fatal(std::string msg){
    // string log_msg = ni_name + ": (Node" + 
    //                     to_string(node.id)+"): "+ msg;
    SC_REPORT_FATAL(LOG_NAME, (msg).c_str());
}