/*******************************************************************************
 * Copyright (C) 2024
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/
/** NetworkInterfaceTlm, which is a specialized version of NetworkInterface

NetworkInterfaceTlm handles the communication between the router and processing element (PE) in the NoC system. 
It manages the sending and receiving of data packets and the flow control between these components. */

#pragma once
#ifndef SC_INCLUDE_DYNAMIC_PROCESSES
#define SC_INCLUDE_DYNAMIC_PROCESSES
#endif

#include "systemc.h"
#include "tlm.h"
#include <tlm_utils/simple_initiator_socket.h> 
#include <tlm_utils/simple_target_socket.h> 
#include <queue>
#include <algorithm>
#include <functional>

#include <iostream> 
#include <vector> 
#include <string> 
#include <cstdlib> 
#include <ctime>    // For time() 

#include "ratatoskrUtils/utils/Structures.h"
#include "ratatoskrUtils/container/PacketContainer.h"
#include "ratatoskrUtils/utils/TrafficTracer.h"
#include "ratatoskrUtils/utils/GlobalResources.h" ///
#include "ratatoskrUtils/networkInterface/NetworkInterface.h"
#include "utils/configuration.h"  

//#include "noc/noc.h"
#include "utils/noc_logger.h"

class NetworkInterfaceTlm : public NetworkInterface {
public:

    tlm_utils::simple_initiator_socket<NetworkInterfaceTlm> socket; //from core
    tlm_utils::simple_target_socket<NetworkInterfaceTlm> target_socket; //from core
    unsigned int source_router_id; //from core
    unsigned int source_core_id; //from core

    //typedef simple_initiator_socket_tagged<NetworkInterfaceTlm> ni_init_socket;
    //typedef simple_target_socket_tagged<NetworkInterfaceTlm> ni_targ_socket;

    std::queue<Packet*> packet_send_queue; //This is a queue where the network interface stores packets that need to be sent.
    std::queue<Packet*> packet_recv_queue; //This is a queue for storing received packets, waiting to be processed.

   
    std::string ni_name;
    sc_in<bool> clk;
    int credit_counter;
    bool resp_in_progress;

    
    PacketPortContainer* packetPortContainer; //This manages packet-level communication with the processing element (PE).
    int creditCounter; //This keeps track of the available credits (flow control units) that manage how much data the network interface can send to the router.
    int lastReceivedCreditID; //This records the last received credit ID from the router to ensure proper flow control and avoid overloading the network.
    TrafficTracer& trafficTracer = TrafficTracer::getInstance(); //A reference to a global traffic tracing system. This helps log and analyze the flow of data (packets and flits) through the network interface for performance monitoring.

    SC_HAS_PROCESS(NetworkInterfaceTlm);

    NetworkInterfaceTlm(sc_module_name nm, Node& node);

    ~NetworkInterfaceTlm() override;


//Overridden Functions: These functions are inherited from the base NetworkInterface class and are specialized for this NetworkInterfaceVC class:
    void initialize() override; //Initializes the network interface. This might involve setting initial values, like clearing the send and receive queues.

    void bind(Connection* conn, SignalContainer* sigContIn,
                                SignalContainer* sigContOut) override; //This binds the network interface to other components (connections, signals) in the system.

    void thread() override; //A process that will run in the SystemC simulation. It controls the main operations of the network interface, like handling packet transmission and reception.

    void send_packet_to_router();
    
    void receivePacketFromPE() override; //This function handles the reception of packets from the processing element (PE). These packets are placed in the packet_send_queue to be sent over the network.

    void receiveFlitFromRouter() override; //Not implemented, another TLM function is used insted. Originally, This function deals with receiving flits from the router and reconstructing them into packets that can be processed by the PE.

    void generateFlitsForPacket(Packet *p) override; //Converts a packet into flits to be sent over the network to the router.

    void getDestination(Packet* p, std::string pos[2]);
    
    void receive_and_process_flit(tlm::tlm_generic_payload& trans);

    /// Forwarding function for receiving data from router
    tlm::tlm_sync_enum nb_transport_fw(tlm::tlm_generic_payload& trans, tlm::tlm_phase& phase, sc_time& delay);

    void log_info(std::string msg);
    void log_error(std::string msg);
    void log_fatal(std::string msg);

};

