/*******************************************************************************
 * Copyright (C) 2024
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/

#include "NetworkManager.h"
#include "utils/noc_logger.h"

NetworkManager::NetworkManager(sc_module_name nm, std::string configFolder){
    setup_logger();
    dbid = rep.registerElement("NetworkManager", 0);
    tlmNoc = new NocTlm("NocTlm", configFolder);

    networkParticipants.resize(globalResources.nodes.size());
    createTrafficPool();
    createClocks();
    createNetworkParticipants();
    //createLinks();
    runNoC();
}

void NetworkManager::createClocks() {
    clocks.resize(globalResources.nodeTypes.size());
    for (const auto &nodeType : globalResources.nodeTypes) {
        clocks.at(nodeType->id) = std::make_unique<sc_clock>(
                ("NodeType" + std::to_string(nodeType->id) + "Clock").c_str(),
                nodeType->clockDelay, SC_NS);
    }
}


void NetworkManager::createTrafficPool() {
    unsigned long numOfPEs = CORE_NO * globalResources.nodes.size() / 2;
#ifndef ENABLE_NETRACE
    if (globalResources.benchmark == "task") {
        tp = std::make_unique<TaskPool>();
    } else if (globalResources.benchmark == "synthetic") {
        tp = std::make_unique<SyntheticPool>();
    } else {
        FATAL("Please specify correct benchmark type");
    }
#endif
// #ifdef ENABLE_NETRACE
//     tp = std::make_unique<NetracePool>("NetracePool");
// #endif
    tp->processingElements.resize(numOfPEs);
}

void NetworkManager::createNetworkParticipants() {
    int numOfPEs = tp->processingElements.size();
    for (Node &n : globalResources.nodes) {
        if (n.type->model == "Router") {
            int id = n.id;

            for (int j = 0; j < CORE_NO; ++j) {
                int index = tlmNoc->cores[id][j]->source_core_id % numOfPEs;

                NetworkInterfaceTlm* ni = tlmNoc->cores[id][j];
                if (!ni) continue; // Ensure ni is valid

                std::string pe_name = "pe_" + std::to_string(n.id) + "_" + std::to_string(j);
                ProcessingElementVC *pe = new ProcessingElementVC(pe_name.c_str(), n, tp.get(), index);

                auto sig1 = std::make_unique<PacketSignalContainer>(
                    ("packetSigCon1_" + std::to_string(n.id) + "_" + std::to_string(j)).c_str());
                auto sig2 = std::make_unique<PacketSignalContainer>(
                    ("packetSigCon2_" + std::to_string(n.id) + "_" + std::to_string(j)).c_str());

                ni->bind(nullptr, sig1.get(), sig2.get());
                pe->bind(nullptr, sig2.get(), sig1.get());
                ni->clk(*clocks.at(n.type->id));

                networkParticipants.push_back(dynamic_cast<NetworkParticipant *>(pe));
                packetSignalContainers.push_back(move(sig1));
                packetSignalContainers.push_back(move(sig2));

                if (index < tp->processingElements.size()) {
                    tp->processingElements[index] = pe;
                }
            }
        }
    }
}


// void NetworkManager::createNetworkParticipants() {
//     int numOfPEs = tp->processingElements.size();
//     for (Node &n : globalResources.nodes) {
//         if (n.type->model == "RouterVC") {
//             int id = n.id;

//             for (int j = 0; j < CORE_NO; ++j) {  // Core index within each router
//                 NetworkInterfaceTlm* ni = tlmNoc->cores[id][j];

//                 std::string pe_name = "pe_" + std::to_string(n.id);
//                 ProcessingElementVC *pe = new ProcessingElementVC(pe_name.c_str(), n, tp.get());
//                 std::unique_ptr<PacketSignalContainer> sig1 = 
//                         std::make_unique<PacketSignalContainer>(
//                         ("packetSigCon1_" + std::to_string(n.id)).c_str());
//                 std::unique_ptr<PacketSignalContainer> sig2 = 
//                         std::make_unique<PacketSignalContainer>(
//                         ("packetSigCon2_" + std::to_string(n.id)).c_str());
//                 ni->bind(nullptr, sig1.get(), sig2.get());
//                 pe->bind(nullptr, sig2.get(), sig1.get());
//                 networkParticipants.push_back(dynamic_cast<NetworkParticipant *>(pe));
//                 packetSignalContainers.push_back(move(sig1));
//                 packetSignalContainers.push_back(move(sig2));
//                 tp->processingElements.at(cores[id][j]->source_core_id % numOfPEs) = pe;
//             }
//         }
//     }
// }


/*
void NetworkManager::createLinks() {
    int link_id = 0;
    for (auto &c : globalResources.connections) {
        int connNodesSize = c.nodes.size();
        if (connNodesSize == 2) { //might extend to bus architecture
            // gets nodes for connections
            Node &node1 = globalResources.nodes.at(c.nodes.at(0));
            Node &node2 = globalResources.nodes.at(c.nodes.at(1));
            bool is_node1_pe = node1.type->model == "ProcessingElement";
            bool is_node2_pe = node2.type->model == "ProcessingElement";
            if (is_node1_pe || is_node2_pe){
                int ni_id = is_node1_pe ? node1.id : node2.id;
                int rout_id = is_node1_pe ? node2.id : node1.id;
                NetworkInterfaceTlm* ni = 
                                dynamic_cast<NetworkInterfaceTlm *>(
                                    networkParticipants.at(ni_id));
                MyRouter* router = tlmNoc->getRouterNodeId(rout_id);
                ni->initiator.bind(*router->target_socket[DIR::Local]);
                router->init_socket[DIR::Local]->bind(ni->target);
            }
            link_id += 2;
        } else {
            LOG(true,"Unsupported number of endpoints in connection "+to_string(c.id));
         }
    }
}
*/
void NetworkManager::runNoC() {
    for (auto &r : networkParticipants) {
        if (r) {
            r->initialize();
        }
    }
    tp->start();
}

NetworkManager::~NetworkManager() {
    for (auto &r : networkParticipants)
        delete r;
    networkParticipants.clear();
}

void NetworkManager::log_info(std::string msg){
    SC_REPORT_INFO(NM_LOG, (msg).c_str());
}