/*******************************************************************************
 * Copyright (C) 2024
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 ******************************************************************************/

// NocTlm.cpp
#include "noc.h"
#include "utils/noc_logger.h"
#include "utils/utils.h"

NocTlm::NocTlm(sc_module_name nm, std::string config_folder): sc_module(nm) {
    setup_logger();
    // Create a semaphore manager instance
    sem_mgr = new semaphore_manager("SemaphoreManager");
    initializeGlobalResources(config_folder);
    setup_routers_and_cores();
    sc_report_handler::set_actions(N_LOG, SC_INFO, SC_LOG|SC_DISPLAY);

}

NocTlm::~NocTlm() {
    // Clean up memory to avoid leaks
    for (int i = 0; i < ROUTER_NO; ++i) {
        for (int j = 0; j < CORE_NO; ++j) {
            delete cores[i][j];  // Delete each core (NI)
        }
        delete routers[i];  // Delete each router
    }
}

void NocTlm::initializeGlobalResources(std::string config_folder) {
    // initialize global resources and rep
    globalResources = GlobalResources::getInstance();
    GlobalReport &globalReport = GlobalReport::getInstance();
    Report &rep = Report::getInstance();
    sleep(1);

    std::string config_path = "config/"+config_folder+"/config.xml";//"config/config.xml";
    globalResources.readConfigFile(config_path);
    globalReport.readConfigFile(config_path);

    std::string network_path = "config/"+config_folder+"/net.xml";
    //globalResources.noc_file;
    globalResources.readNoCLayout(network_path);
    globalResources.readTaskAndMapFiles(globalResources.data_file, globalResources.map_file);
    //globalReport.resizeMatrices();

    globalResources.activateFlitTracing = false;
    globalResources.outputDirectory = "out";
}

void NocTlm::setup_routers_and_cores() {
    // Instantiate ROUTER_NO routers
    //globalResources.nodes.size()
    for (Node &n : globalResources.nodes) {
        if (n.type->model == "Router"){
            int id = n.id;
            std::string router_name = "router" + std::to_string(id);
            routers[id] = new router(router_name.c_str());

            // Store semaphores and out_fifos in vectors
            router_sems.push_back(&routers[id]->sem);
            router_out_fifos.push_back(&routers[id]->out_fifo);
            routers[id]->set_semaphore_manager(sem_mgr);  // Pass semaphore manager to each router
        }
    }

    // Set up pointers to other routers' semaphores and out_fifos for each router
    for (Node &n : globalResources.nodes) {
        if (n.type->model == "Router"){
            int id = n.id;
            routers[id]->other_sems = router_sems;
            routers[id]->other_out_fifos = router_out_fifos;
        }
    }

    // Instantiate ROUTER_NO * CORE_NO cores (CORE_NO per router)
    for (Node &n : globalResources.nodes) {
        if (n.type->model == "Router"){
            int id = n.id;
            for (int j = 0; j < CORE_NO; ++j) {  // Core index within each router
                std::string core_name = "core" + std::to_string(id) + "_" + std::to_string(j);
                cores[id][j] = new NetworkInterfaceTlm(core_name.c_str(), n);
                cores[id][j]->source_router_id = id;  // Set source router ID
                cores[id][j]->source_core_id = CORE_NO*id + j;  // Set source core ID
            }
        }
    }

    // Bind cores to their respective routers
    for (Node &n : globalResources.nodes) {
        if (n.type->model == "Router"){
            int id = n.id;
            for (int j = 0; j < CORE_NO; ++j) {  // Core index within each router
                cores[id][j]->socket.bind(routers[id]->sockets[j]);  // Core's socket to router's target socket
                routers[id]->core_sockets[j].bind(cores[id][j]->target_socket);  // Router's initiator socket to core's target socket
                routers[id]->current_router_id = id;
            }
        }
    }
}

void NocTlm::log_info(std::string msg){
    SC_REPORT_INFO(N_LOG, (msg).c_str());
}

void NocTlm::log_warn(std::string msg){
    SC_REPORT_WARNING(N_LOG, (msg).c_str());
}

void NocTlm::log_error(std::string msg){
    SC_REPORT_ERROR(N_LOG, (msg).c_str());
}